unit CocoAGI;
{$INCLUDE CocoCD.inc}

interface

uses
  CocoDefs, StringTools, CocoSwitch, IniFiles, Classes;

const
  VERSION_NUMBER_STR = 'Version Number';
  LAST_BUILD_STR = 'Last Build';

type
  TCocoAGI = class(TObject)
  private
    fFileName : string;

    fMajorVersion: integer;
    fBuild: integer;
    fMinorVersion: integer;
    fRelease: integer;
    fRegistrationPalette: string;
    fLastBuild: TDateTime;
    fVersionInfo: TStringList;
    fOutputDirectory: string;
    fHashSecondary: string;
    fFrameParser: string;
    fStringInvalid: string;
    fFrameTestDFM: string;
    fFrameConsole: string;
    fMessageShowLL1StartSucc: boolean;
    fFrameAncestor: string;
    fFrameTestProject: string;
    fStringExpected: string;
    fStringNotExpected: string;
    fHashPrimary: string;
    fFrameTestForm: string;
    fMessageShowWarnings: boolean;
    fSwitches: TCocoSwitches;
    fOnShowMessage: TCocoShowMessage;
    fStringVersionFormat: string;
    fStringLastBuildFormat: string;
    function GetVersionNumber: string;
    procedure ReadFrameInformation(const IniFile: TIniFile);
    procedure ReadSwitchInformation(const IniFile: TIniFile);
    procedure ReadVersionInformation(const IniFile: TIniFile);
    procedure ReadMiscInformation(const IniFile: TIniFile);
    procedure ReadStringInformation(const IniFile: TIniFile);

    procedure SetHashPrimary(const Value: string);
    procedure SetHashSecondary(const Value: string);
    procedure SetRegistrationPalette(const Value: string);
    function GetVersionInfoText: string;
    function GetVersionInfoCode: string;
  public
    constructor Create;
    destructor Destroy; override;

    procedure Clear;
    procedure LoadFromFile; overload;
    procedure LoadFromFile(const aFileName : string); overload;
    function UnAliasDirectory(const Dir: string): string;

    property FileName : string read fFileName write fFileName;
    property Switches : TCocoSwitches read fSwitches write fSwitches;

    { MISCELLANEOUS Information }
    property RegistrationPalette : string
        read fRegistrationPalette write SetRegistrationPalette;
    property OutputDirectory : string read fOutputDirectory write fOutputDirectory;

    { VERSION Information }
    property VersionNumber : string read GetVersionNumber;
    property LastBuild : TDateTime read fLastBuild write fLastBuild;
    property VersionInfo : TStringList read fVersionInfo write fVersionInfo;
    property VersionInfoText : string read GetVersionInfoText;
    property VersionInfoCode : string read GetVersionInfoCode;

    property MajorVersion : integer read fMajorVersion write fMajorVersion;
    property MinorVersion : integer read fMinorVersion write fMinorVersion;
    property Release : integer read fRelease write fRelease;
    property Build : integer read fBuild write fBuild;

    { FRAME Information }
    property FrameAncestor : string read fFrameAncestor write fFrameAncestor;
    property FrameParser : string read fFrameParser write fFrameParser;
    property FrameTestProject : string read fFrameTestProject write fFrameTestProject;
    property FrameTestForm : string read fFrameTestForm write fFrameTestForm;
    property FrameTestDFM : string read fFrameTestDFM write fFrameTestDFM;
    property FrameConsole : string read fFrameConsole write fFrameConsole;

    { MESSAGE Information }
    property MessageShowWarnings : boolean read fMessageShowWarnings write fMessageShowWarnings;
    property MessageShowLL1StartSucc : boolean read fMessageShowLL1StartSucc write fMessageShowLL1StartSucc;

    { HASH Information }
    property HashPrimary : string read fHashPrimary write SetHashPrimary;
    property HashSecondary : string read fHashSecondary write SetHashSecondary;

    { String Information }
    property StringExpected : string read fStringExpected write fStringExpected;
    property StringInvalid : string read fStringInvalid write fStringInvalid;
    property StringNotExpected : string read fStringNotExpected write fStringNotExpected;
    property StringLastBuildFormat : string read fStringLastBuildFormat write fStringLastBuildFormat;
    property StringVersionFormat : string read fStringVersionFormat write fStringVersionFormat;

    { Events }
    property OnShowMessage : TCocoShowMessage read fOnShowMessage write fOnShowMessage;
  end; {TCocoAGI}

implementation

uses
   SysUtils, Forms;

const
  HomeDirAlias = '$(CocoR)';
  Expected_Msg = '%D% expected';
  Invalid_Msg = 'invalid %D%';
  Not_Expected_Msg = 'this symbol not expected in %D%';
   
{ TCocoAGI }

procedure TCocoAGI.Clear;
begin
  fFileName := '';

  fVersionInfo.Clear;
  fMajorVersion := 0;
  fBuild := 0;
  fMinorVersion := 0;
  fRelease := 0;
  fLastBuild := 0;

  RegistrationPalette := '';
  fOutputDirectory := '';

  HashSecondary := '';
  HashPrimary := '';

  fMessageShowLL1StartSucc := TRUE;
  fMessageShowWarnings := TRUE;
  fStringLastBuildFormat := '';
  fStringVersionFormat := '';
end; {Clear}

constructor TCocoAGI.Create;
begin
  fVersionInfo := TStringList.Create;
  Clear;
end; {Create}

destructor TCocoAGI.Destroy;
begin
  if Assigned(fVersionInfo) then
    FreeAndNil(fVersionInfo);
  inherited;
end; {Destroy}

function TCocoAGI.GetVersionNumber: string;
begin
  Result := '';
end; {GetVersionNumber}

function TCocoAGI.UnAliasDirectory(const Dir: string): string;
var
  PathFollowingAlias : string;
begin
  if Uppercase(copy(Dir,1,length(HomeDirAlias))) = UpperCase(HomeDirAlias) then
    begin
      PathFollowingAlias := copy(Dir,length(HomeDirAlias) + 1,MaxInt);
      while (Length(PathFollowingAlias) > 0) AND (PathFollowingAlias[1] = SysUtils.PathDelim) do
        Delete(PathFollowingAlias, 1, 1);
      Result := IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName))
          + PathFollowingAlias;
    end
  else
    Result := Dir;
end; {UnAliasDirectory}

procedure TCocoAGI.ReadFrameInformation(const IniFile : TIniFile);
begin
  fFrameAncestor := UnAliasDirectory(IniFile.ReadString('Frames', 'Ancestor', ''));
  fFrameParser := UnAliasDirectory(IniFile.ReadString('Frames', 'Parser', ''));
  fFrameTestProject := UnAliasDirectory(IniFile.ReadString('Frames', 'Test Project', ''));
  fFrameTestForm := UnAliasDirectory(IniFile.ReadString('Frames', 'Test Form', ''));
  fFrameTestDFM := UnAliasDirectory(IniFile.ReadString('Frames', 'Test DFM', ''));
  fFrameConsole := UnAliasDirectory(IniFile.ReadString('Frames', 'Console', ''));
end; {ReadFrameInformation}

procedure TCocoAGI.ReadStringInformation(const IniFile : TIniFile);
begin
  fStringExpected := IniFile.ReadString('Error Msg', 'Expected', Expected_Msg);
  fStringInvalid := IniFile.ReadString('Error Msg', 'Invalid', Invalid_Msg);
  fStringNotExpected := IniFile.ReadString('Error Msg', 'Not Expected', Not_Expected_Msg);
  fStringLastBuildFormat := IniFile.ReadString('Strings', 'Last Build Format', '');
  fStringVersionFormat := IniFile.ReadString('Strings', 'Version Format', '');
end; {ReadStringInformation}

procedure TCocoAGI.ReadMiscInformation(const IniFile: TIniFile);
begin
  fOutputDirectory := UnAliasDirectory(IniFile.ReadString('Output', 'Directory', ''));
  RegistrationPalette := IniFile.ReadString('Registration', 'Palette', 'Coco/R');

  fMessageShowWarnings := IniFile.ReadBool('Messages','Show Warnings',TRUE);
  fMessageShowLL1StartSucc := IniFile.ReadBool('Messages','Show LL(1) Start Succ',TRUE);
  HashPrimary := IniFile.ReadString('Hash','Primary','');
  HashSecondary := IniFile.ReadString('Hash','Secondary','');
end; {ReadMiscInformation}

procedure TCocoAGI.ReadSwitchInformation(const IniFile: TIniFile);
var
  Switch : TCocoSwitch;
  SwitchChar : string;
begin
  fSwitches.Clear;
  for Switch := low(TCocoSwitch) to High(TCocoSwitch) do
  begin
    SwitchChar := fSwitches.SwitchChar(Switch);
    if IniFile.ReadBool('Options', SwitchChar, Switch IN [crsSpaceAsWhitespace]) then
      fSwitches.SwitchSet := fSwitches.SwitchSet + [Switch];
  end;
end; {ReadSwitchInformation}

procedure TCocoAGI.ReadVersionInformation(const IniFile: TIniFile);
var
  TempStr : string;
  VersionSection : TStringList;
  i : integer;
  LastBuildStr : string;
begin
  TempStr := IniFile.ReadString('Version Info', 'Version Number', '');
  if TempStr = '' then
  begin
    fMajorVersion := IniFile.ReadInteger('Version Info', 'Major Version', 0);
    IniFile.DeleteKey('Version Info','Major Version');
    fMinorVersion := IniFile.ReadInteger('Version Info', 'Minor Version', 0);
    IniFile.DeleteKey('Version Info','Minor Version');
    fRelease := IniFile.ReadInteger('Version Info', 'Release', 0);
    IniFile.DeleteKey('Version Info','Release');
    fBuild := IniFile.ReadInteger('Version Info', 'Next Build', 0);
    IniFile.DeleteKey('Version Info','Next Build');
  end
  else
  begin
    fMajorVersion := StrToIntDef(StrTok(TempStr,'.'),0);
    fMinorVersion := StrToIntDef(StrTok(TempStr,'.'),0);
    fRelease := StrToIntDef(StrTok(TempStr,'.'),0);
    fBuild := StrToIntDef(TempStr,0);
  end;
  LastBuildStr := IniFile.ReadString('Version Info', 'Last Build', '');
  fLastBuild := StrToDateTimeDef(LastBuildStr, 0);
  fVersionInfo.Clear;
  VersionSection := TStringList.Create;
  try
    IniFile.ReadSection('Version Info', VersionSection);
    for i := 0 to VersionSection.Count - 1 do
      if (UpperCase(VersionSection[i]) <> 'LAST BUILD')
          AND (UpperCase(VersionSection[i]) <> 'VERSION NUMBER') then
        fVersionInfo.Add(VersionSection[i] + '='
            + IniFile.ReadString('Version Info', VersionSection[i], ''));
  finally
    FreeAndNIL(VersionSection);
  end; {try..finally}
end; {ReadVersionInformation}

procedure TCocoAGI.LoadFromFile;
var
  IniFile : TIniFile;
begin
  if FileExists(fFileName) then
  begin
    IniFile := TIniFile.Create(fFileName);
    try
      ReadFrameInformation(IniFile);
      ReadSwitchInformation(IniFile);
      ReadVersionInformation(IniFile);
      ReadMiscInformation(IniFile);
      ReadStringInformation(IniFile);
    finally
      IniFile.Free;
    end;
  end;
end; {LoadFromFile}

procedure TCocoAGI.LoadFromFile(const aFileName: string);
begin
  fFileName := aFileName;
  LoadFromFile;
end; {LoadFromFile}

procedure TCocoAGI.SetHashPrimary(const Value: string);
begin
  fHashPrimary := Value;
  if Assigned(fSwitches) then
    fSwitches.PrimaryHash := fHashPrimary;
end; {SetHashPrimary}

procedure TCocoAGI.SetHashSecondary(const Value: string);
begin
  fHashSecondary := Value;
  if Assigned(fSwitches) then
    fSwitches.SecondaryHash := fHashSecondary;
end; {SetHashSecondary}

procedure TCocoAGI.SetRegistrationPalette(const Value: string);
begin
  fRegistrationPalette := Value;
  if Assigned(fSwitches) then
    fSwitches.RegistrationPalette := fRegistrationPalette;
end; {SetRegistrationPalette}

function TCocoAGI.GetVersionInfoCode: string;
var
  i : integer;
begin
  Result := '';
  for i := 0 to VersionInfo.Count - 1 do
  begin
    Result := Result + AnsiQuotedStr(StringReplace(VersionInfo[i],'=',': ',[rfReplaceAll]),#39);
    if i < VersionInfo.Count - 1 then
      Result := Result + ' + #13#10 +' + CRLF;
  end;
end; {GetVersionInfoText}

function TCocoAGI.GetVersionInfoText: string;
var
  i : integer;
begin
  Result := '';
  for i := 0 to VersionInfo.Count - 1 do
  begin
    Result := Result + StringReplace(VersionInfo[i],'=',': ',[rfReplaceAll]);
    if i < VersionInfo.Count - 1 then
      Result := Result + CRLF;
  end;
end; {GetVersionInfoText}

end.

