unit CocoRegistry;
{$INCLUDE CocoCD.inc}

interface

uses
  Classes,
  Graphics;

type
  TCocoRegistry = class(TObject)
  private
    fDefaultFrameDir: string;
    fHomeDir : string;
    fDefaultErrorMsgExpected : string;
    fDefaultErrorMsgInvalid : string;
    fDefaultErrorMsgNotExpected : string;
    fDefaultVersionFormat: string;
    fDefaultLastBuildFormat: string;

    function GetHomeDir: string;
  public
    constructor Create;
    procedure LoadDefaults;
    procedure SaveDefaults;
    procedure SetHomeDir(const Dir: string);
    function AliasDirectory(const Dir : string) : string;
    function UnAliasDirectory(const Dir : string) : string;

    property DefaultErrorMsgExpected : string read fDefaultErrorMsgExpected write fDefaultErrorMsgExpected;
    property DefaultErrorMsgInvalid : string read fDefaultErrorMsgInvalid write fDefaultErrorMsgInvalid;
    property DefaultErrorMsgNotExpected : string read fDefaultErrorMsgNotExpected write fDefaultErrorMsgNotExpected;
    property DefaultLastBuildFormat : string read fDefaultLastBuildFormat write fDefaultLastBuildFormat;
    property DefaultVersionFormat : string read fDefaultVersionFormat write fDefaultVersionFormat;
    property DefaultFrameDir : string read fDefaultFrameDir write fDefaultFrameDir;
    property HomeDir : string read GetHomeDir;
  end; {TCocoRegistry}

implementation

uses
  StringTools, SysUtils, TZRegistry;

const
  HKEY_LOCAL_MACHINE = $80000002;
  HomeDirAlias = '$(CocoR)';
  BaseRegistry = 'Software\CocoR for Delphi';

  Expected_Msg = '%D% expected';
  Invalid_Msg = 'invalid %D%';
  Not_Expected_Msg = 'this symbol not expected in %D%';

procedure TCocoRegistry.LoadDefaults;
var
  Registry : TTZRegistry;
begin
  Registry := TTZRegistry.Create;
  try
    Registry.OpenKey(BaseRegistry + '\Defaults', TRUE);
    fDefaultFrameDir := UnAliasDirectory(Registry.ReadStringDef('Frame Dir', HomeDirAlias));

    fDefaultErrorMsgExpected := Registry.ReadStringDef(
        'Default Error Msg Expected', Expected_Msg);
    fDefaultErrorMsgInvalid := Registry.ReadStringDef(
        'Default Error Msg Invalid', Invalid_Msg);
    fDefaultErrorMsgNotExpected := Registry.ReadStringDef(
        'Default Error Msg Not Expected', Not_Expected_Msg);
    fDefaultVersionFormat := Registry.ReadStringDef('Default Version Format', '');
    fDefaultLastBuildFormat := Registry.ReadStringDef('Default Last Build Format', '');
  finally
    Registry.Free;
  end;
end;

procedure TCocoRegistry.SaveDefaults;
var
  Registry : TTZRegistry;
begin
  Registry := TTZRegistry.Create;
  try
    Registry.OpenKey(BaseRegistry + '\Defaults', TRUE);
    Registry.WriteString('Frame Dir',AliasDirectory(fDefaultFrameDir));
    Registry.WriteString('Default Error Msg Expected', fDefaultErrorMsgExpected);
    Registry.WriteString('Default Error Msg Invalid', fDefaultErrorMsgInvalid);
    Registry.WriteString('Default Error Msg Not Expected', fDefaultErrorMsgNotExpected);
    Registry.WriteString('Default Version Format', fDefaultVersionFormat);
    Registry.WriteString('Default Last Build Format', fDefaultLastBuildFormat);
  finally
    Registry.Free;
  end;
end;

function TCocoRegistry.GetHomeDir: string;
var
  Reg : TTZRegistry;
begin
  if fHomeDir = '' then
  begin
    Reg := TTZRegistry.Create;
    try
      Reg.RootKey := HKEY_LOCAL_MACHINE;
      Reg.OpenKey(BaseRegistry,TRUE);
      fHomeDir := Reg.ReadString('Path');
    finally
      Reg.Free;
    end;
  end;
  Result := fHomeDir;
end;

procedure TCocoRegistry.SetHomeDir(const Dir : string);
var
  Reg : TTZRegistry;
begin
  Reg := TTZRegistry.Create;
  try
    Reg.RootKey := HKEY_LOCAL_MACHINE;
    Reg.OpenKey(BaseRegistry, TRUE);
    Reg.WriteString('Path', Dir);
  finally
    Reg.Free;
  end;
end;

function TCocoRegistry.AliasDirectory(const Dir: string): string;
begin
  if Uppercase(copy(Dir,1,length(HomeDir))) = UpperCase(HomeDir) then
    Result := HomeDirAlias + copy(Dir,length(HomeDir) + 1,MaxInt)
  else
    Result := Dir;
end; {AliasDirectory}

function TCocoRegistry.UnAliasDirectory(const Dir: string): string;
begin
  if Uppercase(copy(Dir,1,length(HomeDirAlias))) = UpperCase(HomeDirAlias) then
    Result := HomeDir + copy(Dir,length(HomeDirAlias) + 1,MaxInt)
  else
    Result := Dir;
end; {UnAliasDirectory}

constructor TCocoRegistry.Create;
begin
  fHomeDir := '';
end; {Create}

end.

